#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import base64
import logging
from urllib.parse import urlparse, urlunparse
from typing import Iterator, Tuple, Optional, List, Callable, Dict, Iterable, Union

from .commands.helpers.enterprise import user_has_privilege, is_addon_enabled
from .constants import KEEPER_PUBLIC_HOSTS
from . import api, crypto, utils, rest_api, vault
from .proto import breachwatch_pb2, client_pb2, APIRequest_pb2
from .error import KeeperApiError, CommandError
from .params import KeeperParams
from .security_audit import needs_security_audit, update_security_audit_data
from .vault import KeeperRecord


class BreachWatch(object):
    def __init__(self):
        self.rest_api = None
        self.domain_token = None
        self.email_token = None
        self.password_token = None
        self.send_audit_events = False

    @staticmethod
    def extract_password(record):     # type: (vault.KeeperRecord) -> Optional[str]
        if isinstance(record, vault.PasswordRecord):
            return record.password
        if isinstance(record, vault.TypedRecord):
            password_field = record.get_typed_field('password')
            if password_field:
                return password_field.get_default_value(str)

    @staticmethod
    def extract_url(record):     # type: (vault.KeeperRecord) -> Optional[str]
        if isinstance(record, vault.PasswordRecord):
            return record.link
        if isinstance(record, vault.TypedRecord):
            url_field = record.get_typed_field('url')
            if url_field:
                return url_field.get_default_value(str)

    def scan_password(self, params, password, euid=None):
        # type: (KeeperParams, str, Optional[bytes]) -> Optional[breachwatch_pb2.HashStatus]

        bw_hash = utils.breach_watch_hash(password)
        if not euid:
            score = utils.password_score(password)
            if score < 40:
                result = breachwatch_pb2.HashStatus()
                result.hash1 = bw_hash
                result.breachDetected = True
                return result

        self._ensure_init(params)
        check = breachwatch_pb2.HashCheck()
        check.hash1 = bw_hash
        if euid:
            check.euid = euid
        rq = breachwatch_pb2.BreachWatchStatusRequest()
        rq.hashCheck.append(check)
        rs = self._execute_status(rq)
        return rs.hashStatus[0]

    def scan_passwords(self, params, passwords):
        # type: (KeeperParams, Iterator[str]) -> Iterator[Tuple[str, breachwatch_pb2.HashStatus]]
        results = {}      # type: Dict[str, breachwatch_pb2.HashStatus]
        bw_hashes = {}    # type: Dict[bytes, str]
        for password in passwords:
            if isinstance(password, str) and len(password) > 0:
                score = utils.password_score(password)
                bw_hash = utils.breach_watch_hash(password)
                if score >= 40:
                    bw_hashes[bw_hash] = password
                else:
                    status = breachwatch_pb2.HashStatus()
                    status.hash1 = bw_hash
                    status.breachDetected = True
                    results[password] = status
        if len(bw_hashes) > 0:
            logging.info('Breachwatch: %d passwords to scan', len(bw_hashes))
            hashes = []     # type: List[breachwatch_pb2.HashCheck]
            for bw_hash in bw_hashes:
                check = breachwatch_pb2.HashCheck()
                check.hash1 = bw_hash
                hashes.append(check)
            self._ensure_init(params)

            while len(hashes) > 0:
                chunk = hashes[:500]
                hashes = hashes[500:]

                rq = breachwatch_pb2.BreachWatchStatusRequest()
                rq.hashCheck.extend(chunk)

                rs = self._execute_status(rq)
                for status in rs.hashStatus:
                    password = bw_hashes.get(status.hash1)
                    if isinstance(password, str) and len(password) > 0:
                        results[password] = status

        for password in results:
            yield password, results[password]

    def scan_and_store_record_status(self, params, record, force_update=False):
        # type: (KeeperParams, KeeperRecord, Optional[bool]) -> None
        def get_euid():
            result = None
            if bw_record:
                bw_pw_objs = bw_record.get('data_unencrypted', {}).get('passwords', [])
                euids = [x.get('euid') for x in bw_pw_objs if x.get('euid')]
                if euids:
                    result = base64.b64decode(next(iter(euids)))
            return result

        def get_last_pw():
            result = ''
            if bw_record:
                bw_pw_objs = bw_record.get('data_unencrypted', {}).get('passwords', [])
                passwords = [x.get('value') for x in bw_pw_objs if x.get('value')]
                if passwords:
                    result = next(iter(passwords))
            return result

        def update_bw_data():
            result = None
            bwrq = breachwatch_pb2.BreachWatchRecordRequest()
            bwrq.recordUid = utils.base64_url_decode(record_uid)
            bwrq.breachWatchInfoType = breachwatch_pb2.RECORD
            bwrq.updateUserWhoScanned = True
            bw_data = client_pb2.BreachWatchData()

            if record_password:
                hash_status = self.scan_password(params, record_password, get_euid())

                if hash_status.breachDetected:
                    logging.info('High-Risk password detected')
                    if self.send_audit_events:
                        params.queue_audit_event('bw_record_high_risk')
                bw_password = client_pb2.BWPassword()
                bw_password.value = record_password
                bw_password.status = client_pb2.WEAK if hash_status.breachDetected else client_pb2.GOOD
                bw_password.euid = hash_status.euid
                bw_data.passwords.append(bw_password)
                result = bw_password.status
            try:
                data = bw_data.SerializeToString()
                record_key = params.record_cache[record_uid]['record_key_unencrypted']
                bwrq.encryptedData = crypto.encrypt_aes_v2(data, record_key)
                rq = breachwatch_pb2.BreachWatchUpdateRequest()
                rq.breachWatchRecordRequest.append(bwrq)
                rs = api.communicate_rest(params, rq, 'breachwatch/update_record_data',
                                          rs_type=breachwatch_pb2.BreachWatchUpdateResponse)
                status = rs.breachWatchRecordStatus[0]
                if status.reason:
                    raise Exception(status.reason)
            except Exception as e:
                logging.warning('BreachWatch: %s', str(e))
            return result

        def skip_update():
            if record_password == get_last_pw():
                return True
            return False

        record_uid = record.record_uid
        bw_record = params.breach_watch_records.get(record_uid) if params.breach_watch_records else None

        record_password = BreachWatch.extract_password(record) or ''
        if skip_update():
            return None

        bw_res = update_bw_data()
        if not record_password:
            euid = get_euid()
            if euid:
                params.breach_watch.delete_euids(params, [euid])
        api.sync_down(params)
        return bw_res

    @staticmethod
    def update_security_data(params, record, force_update=False):
        # type: (KeeperParams, KeeperRecord, Optional[bool]) -> None
        if force_update or needs_security_audit(params, record):
            update_security_audit_data(params, [record])

    @staticmethod
    def save_reused_pw_count(params):
        def get_reused_pw_count(recs):   # type: (Iterable[vault.KeeperRecord]) -> int
            pw_counts = {}
            for rec in recs:
                pw = BreachWatch.extract_password(rec)
                if pw:
                    pw_count = pw_counts.get(pw, 0)
                    pw_counts[pw] = pw_count + 1

            dupe_pw_counts = {k: v for k, v in pw_counts.items() if v > 1}
            return sum([count for count in dupe_pw_counts.values()])

        if params.enterprise_ec_key:
            api.sync_down(params)
            owned = [uid for uid, own in params.record_owner_cache.items()
                              if own.owner is True and uid in params.record_cache]
            owned_recs = [x for x in (vault.KeeperRecord.load(params, ruid) for ruid in owned)
                          if x and x.version in (2, 3)]
            total_reused = get_reused_pw_count(owned_recs)
            save_rq = APIRequest_pb2.ReusedPasswordsRequest()
            save_rq.count = total_reused
            api.communicate_rest(params, save_rq, 'enterprise/set_reused_passwords')

    def delete_euids(self, params, euids):
        self._ensure_init(params)
        while euids:
            chunk = euids[:999]
            euids = euids[999:]
            rq = breachwatch_pb2.BreachWatchStatusRequest()
            rq.removedEuid.extend(chunk)
            self._execute_status(rq)

    def _execute_status(self, rq):
        rq.anonymizedToken = self.password_token
        api_request_payload = APIRequest_pb2.ApiRequestPayload()
        api_request_payload.payload = rq.SerializeToString()
        rs = rest_api.execute_rest(self.rest_api, 'breachwatch/status', api_request_payload)
        if isinstance(rs, bytes):
            bw_rs = breachwatch_pb2.BreachWatchStatusResponse()
            bw_rs.ParseFromString(rs)
            return bw_rs
        elif isinstance(rs, dict):
            raise KeeperApiError(rs['error'], rs['message'])
        raise KeeperApiError('Error', 'breachwatch/status')

    def _ensure_init(self, params):
        if self.rest_api:
            if not self.password_token:
                raise KeeperApiError('not_initialized', 'BreachWatch init error.')
            return

        url_comp = urlparse(params.rest_context.server_base)
        us_server = KEEPER_PUBLIC_HOSTS['US']
        if not url_comp.netloc.endswith(us_server):
            for region in KEEPER_PUBLIC_HOSTS.values():
                if url_comp.netloc.endswith(region):
                    url_comp = url_comp._replace(netloc=url_comp.netloc[0: -len(region)] + us_server)
                    break
        bw_endpoint = rest_api.RestApiContext(server=urlunparse(url_comp), locale=params.rest_context.locale)
        bw_endpoint.server_key_id = params.rest_context.server_key_id
        self.rest_api = bw_endpoint

        rs = api.communicate_rest(params, None, 'breachwatch/initialize',
                                  rs_type=breachwatch_pb2.BreachWatchTokenResponse)
        if rs.clientEncrypted:
            enc_token = rs.breachWatchToken
            breach_watch_token = crypto.decrypt_aes_v2(enc_token, params.data_key)
        else:
            breach_watch_token = rs.breachWatchToken
            enc_token = crypto.encrypt_aes_v2(breach_watch_token, params.data_key)
            rq = breachwatch_pb2.BreachWatchTokenRequest()
            rq.breachWatchToken = enc_token
            api.communicate_rest(params, rq, 'breachwatch/save_token')

        rq = breachwatch_pb2.BreachWatchTokenRequest()
        rq.breachWatchToken = breach_watch_token
        rs = api.communicate_rest(params, rq, 'breachwatch/anonymize_token',
                                  rs_type=breachwatch_pb2.AnonymizedTokenResponse)
        self.domain_token = rs.domainToken
        self.email_token = rs.emailToken
        self.password_token = rs.passwordToken

        if params.enforcements:
            pass

    @staticmethod
    def get_record_status(params, record_uid):  # type: (KeeperParams, str) -> Optional[dict]
        if not params.breach_watch_records:
            return
        bw_record = params.breach_watch_records.get(record_uid)
        if bw_record:
            data_obj = bw_record.get('data_unencrypted')
            if data_obj and 'passwords' in data_obj:
                record = vault.KeeperRecord.load(params, record_uid)
                if record:
                    record_password = BreachWatch.extract_password(record)
                    if record_password:
                        return next((x for x in data_obj['passwords'] if x.get('value', '') == record_password), None)

    @staticmethod
    def get_records(params,            # type: KeeperParams
                    callback,          # type: Callable[[vault.KeeperRecord, Optional[dict]], bool]
                    owned=False        # type: bool
                    ):                 # type: (...) -> Iterator[Tuple[vault.KeeperRecord, Optional[dict]]]
        if not params.record_cache:
            return

        for record_uid in params.record_cache:
            record = vault.KeeperRecord.load(params, record_uid)
            if not record:
                continue
            if owned:
                if record_uid not in params.record_owner_cache:
                    continue
                if not params.record_owner_cache[record_uid].owner is True:
                    continue

            password = BreachWatch.extract_password(record)
            if not password:
                continue

            if isinstance(password, str) and password:
                password_dict = None
                if params.breach_watch_records:
                    bwr = params.breach_watch_records.get(record_uid)
                    data_obj = bwr.get('data_unencrypted') if bwr else None
                    if data_obj and 'passwords' in data_obj:
                        password_dict = next((x for x in data_obj['passwords'] if x.get('value', '') == password), None)
                if callback(record, password_dict):
                    yield record, password_dict

    @staticmethod
    def get_records_to_scan(params):  # type: (KeeperParams) -> Iterator[Tuple[vault.KeeperRecord, Optional[dict]]]
        yield from BreachWatch.get_records(params, lambda r, b: b is None, owned=True)

    @staticmethod
    def check_status(bwr, statuses):
        if isinstance(bwr, dict) and isinstance(statuses, set):
            return bwr.get('status', '').casefold() in statuses
        return False

    @staticmethod
    def get_records_by_status(params, status, owned=False):
        # type: (KeeperParams, Optional[str, List[str]], bool) -> Iterator[Tuple[vault.KeeperRecord, Optional[dict]]]
        statuses = set()
        if status:
            if isinstance(status, list):
                statuses.update((x.casefold() for x in status if isinstance(x, str)))
            elif isinstance(status, str):
                statuses.add(status.casefold())
        else:
            statuses.update((x.casefold() for x in client_pb2.BWStatus.keys()))

        yield from params.breach_watch.get_records(params, lambda r, b: BreachWatch.check_status(b, statuses), owned)


    @staticmethod
    def scan_and_update_security_data(params, record_uid, bw_obj=None, force_update=False, set_reused_pws=True):
        # type: (KeeperParams, Union[str, List[str]], Optional[BreachWatch], Optional[bool], Optional[bool]) -> None
        api.sync_down(params)
        record = vault.KeeperRecord.load(params, record_uid)
        if not record:
            return
        if not isinstance(record, (vault.PasswordRecord, vault.TypedRecord)):
            return

        if params.breach_watch:
            params.breach_watch.scan_and_store_record_status(params, record)
        BreachWatch.update_security_data(params, record, force_update=force_update)
        if set_reused_pws:
            BreachWatch.save_reused_pw_count(params)
        api.sync_down(params)

    @staticmethod
    def validate_reporting(cmd, params):
        msg_no_priv = 'You do not have the required privilege to run a BreachWatch report'
        msg_no_addon = ('BreachWatch is not enabled for this enterprise. '
                        'Please visit https://www.keepersecurity.com/breachwatch.html for more information.')

        privilege = 'run_reports'
        addon = 'enterprise_breach_watch'
        error_msg = msg_no_priv if not user_has_privilege(params, privilege) \
            else msg_no_addon if not is_addon_enabled(params, addon) \
            else None
        if error_msg:
            raise CommandError(cmd, error_msg)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pam.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import enterprise_pb2 as enterprise__pb2
from . import record_pb2 as record__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tpam.proto\x12\x03PAM\x1a\x10\x65nterprise.proto\x1a\x0crecord.proto\"\x83\x01\n\x13PAMRotationSchedule\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x18\n\x10\x63onfigurationUid\x18\x02 \x01(\x0c\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x14\n\x0cscheduleData\x18\x04 \x01(\t\x12\x12\n\nnoSchedule\x18\x05 \x01(\x08\"K\n\x1cPAMRotationSchedulesResponse\x12+\n\tschedules\x18\x01 \x03(\x0b\x32\x18.PAM.PAMRotationSchedule\"\x94\x01\n\x13PAMOnlineController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63onnectedOn\x18\x02 \x01(\x03\x12\x11\n\tipAddress\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\t\x12-\n\x0b\x63onnections\x18\x05 \x03(\x0b\x32\x18.PAM.PAMWebRtcConnection\"\xa7\x01\n\x13PAMWebRtcConnection\x12\x15\n\rconnectionUid\x18\x01 \x01(\x0c\x12\'\n\x04type\x18\x02 \x01(\x0e\x32\x19.PAM.WebRtcConnectionType\x12\x11\n\trecordUid\x18\x03 \x01(\x0c\x12\x10\n\x08userName\x18\x04 \x01(\t\x12\x11\n\tstartedOn\x18\x05 \x01(\x03\x12\x18\n\x10\x63onfigurationUid\x18\x06 \x01(\x0c\"Y\n\x14PAMOnlineControllers\x12\x12\n\ndeprecated\x18\x01 \x03(\x0c\x12-\n\x0b\x63ontrollers\x18\x02 \x03(\x0b\x32\x18.PAM.PAMOnlineController\"9\n\x10PAMRotateRequest\x12\x12\n\nrequestUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\"A\n\x16PAMControllersResponse\x12\'\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x12.PAM.PAMController\"=\n\x13PAMRemoveController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\t\"L\n\x1bPAMRemoveControllerResponse\x12-\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x18.PAM.PAMRemoveController\"=\n\x10PAMModifyRequest\x12)\n\noperations\x18\x01 \x03(\x0b\x32\x15.PAM.PAMDataOperation\"\x98\x01\n\x10PAMDataOperation\x12,\n\roperationType\x18\x01 \x01(\x0e\x32\x15.PAM.PAMOperationType\x12\x30\n\rconfiguration\x18\x02 \x01(\x0b\x32\x19.PAM.PAMConfigurationData\x12$\n\x07\x65lement\x18\x03 \x01(\x0b\x32\x13.PAM.PAMElementData\"e\n\x14PAMConfigurationData\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\"E\n\x0ePAMElementData\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12\x11\n\tparentUid\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\"m\n\x19PAMElementOperationResult\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12+\n\x06result\x18\x02 \x01(\x0e\x32\x1b.PAM.PAMOperationResultType\x12\x0f\n\x07message\x18\x03 \x01(\t\"B\n\x0fPAMModifyResult\x12/\n\x07results\x18\x01 \x03(\x0b\x32\x1e.PAM.PAMElementOperationResult\"x\n\nPAMElement\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x03 \x01(\x03\x12\x14\n\x0clastModified\x18\x04 \x01(\x03\x12!\n\x08\x63hildren\x18\x05 \x03(\x0b\x32\x0f.PAM.PAMElement\"#\n\x14PAMGenericUidRequest\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\"%\n\x15PAMGenericUidsRequest\x12\x0c\n\x04uids\x18\x01 \x03(\x0c\"\xab\x01\n\x10PAMConfiguration\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x05 \x01(\x03\x12\x14\n\x0clastModified\x18\x06 \x01(\x03\x12!\n\x08\x63hildren\x18\x07 \x03(\x0b\x32\x0f.PAM.PAMElement\"B\n\x11PAMConfigurations\x12-\n\x0e\x63onfigurations\x18\x01 \x03(\x0b\x32\x15.PAM.PAMConfiguration\"\xff\x01\n\rPAMController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63ontrollerName\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65viceToken\x18\x03 \x01(\t\x12\x12\n\ndeviceName\x18\x04 \x01(\t\x12\x0e\n\x06nodeId\x18\x05 \x01(\x03\x12\x0f\n\x07\x63reated\x18\x06 \x01(\x03\x12\x14\n\x0clastModified\x18\x07 \x01(\x03\x12\x16\n\x0e\x61pplicationUid\x18\x08 \x01(\x0c\x12\x30\n\rappClientType\x18\t \x01(\x0e\x32\x19.Enterprise.AppClientType\x12\x15\n\risInitialized\x18\n \x01(\x08\"%\n\x12\x43ontrollerResponse\x12\x0f\n\x07payload\x18\x01 \x01(\t\"M\n\x1aPAMConfigurationController\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x15\n\rcontrollerUid\x18\x02 \x01(\x0c\"\xa3\x01\n\x17\x43onfigurationAddRequest\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x11\n\trecordKey\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12(\n\x0brecordLinks\x18\x04 \x03(\x0b\x32\x13.Records.RecordLink\x12#\n\x05\x61udit\x18\x05 \x01(\x0b\x32\x14.Records.RecordAudit\"J\n\x10RelayAccessCreds\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\x12\x12\n\nserverTime\x18\x03 \x01(\x03\"\xbf\x01\n\x0cPAMRecording\x12\x15\n\rconnectionUid\x18\x01 \x01(\x0c\x12,\n\rrecordingType\x18\x02 \x01(\x0e\x32\x15.PAM.PAMRecordingType\x12\x11\n\trecordUid\x18\x03 \x01(\x0c\x12\x10\n\x08userName\x18\x04 \x01(\t\x12\x11\n\tstartedOn\x18\x05 \x01(\x03\x12\x0e\n\x06length\x18\x06 \x01(\x05\x12\x10\n\x08\x66ileSize\x18\x07 \x01(\x03\x12\x10\n\x08protocol\x18\x08 \x01(\t\">\n\x15PAMRecordingsResponse\x12%\n\nrecordings\x18\x01 \x03(\x0b\x32\x11.PAM.PAMRecording\"*\n\x07PAMData\x12\x0e\n\x06vertex\x18\x01 \x01(\x0c\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\x0c\"\x17\n\x07UidList\x12\x0c\n\x04uids\x18\x01 \x03(\x0c\"\xd0\x02\n\x11PAMResourceConfig\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x17\n\nnetworkUid\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x12\x15\n\x08\x61\x64minUid\x18\x03 \x01(\x0cH\x01\x88\x01\x01\x12\x11\n\x04meta\x18\x04 \x01(\x0cH\x02\x88\x01\x01\x12\x1f\n\x12\x63onnectionSettings\x18\x05 \x01(\x0cH\x03\x88\x01\x01\x12\'\n\x0c\x63onnectUsers\x18\x06 \x01(\x0b\x32\x0c.PAM.UidListH\x04\x88\x01\x01\x12\x16\n\tdomainUid\x18\x07 \x01(\x0cH\x05\x88\x01\x01\x12\x18\n\x0bjitSettings\x18\x08 \x01(\x0cH\x06\x88\x01\x01\x42\r\n\x0b_networkUidB\x0b\n\t_adminUidB\x07\n\x05_metaB\x15\n\x13_connectionSettingsB\x0f\n\r_connectUsersB\x0c\n\n_domainUidB\x0e\n\x0c_jitSettings*\x8e\x01\n\x14WebRtcConnectionType\x12\x0e\n\nCONNECTION\x10\x00\x12\n\n\x06TUNNEL\x10\x01\x12\x07\n\x03SSH\x10\x02\x12\x07\n\x03RDP\x10\x03\x12\x08\n\x04HTTP\x10\x04\x12\x07\n\x03VNC\x10\x05\x12\n\n\x06TELNET\x10\x06\x12\t\n\x05MYSQL\x10\x07\x12\x0e\n\nSQL_SERVER\x10\x08\x12\x0e\n\nPOSTGRESQL\x10\t*@\n\x10PAMOperationType\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\n\n\x06UPDATE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03*p\n\x16PAMOperationResultType\x12\x0f\n\x0bPOT_SUCCESS\x10\x00\x12\x15\n\x11POT_UNKNOWN_ERROR\x10\x01\x12\x16\n\x12POT_ALREADY_EXISTS\x10\x02\x12\x16\n\x12POT_DOES_NOT_EXIST\x10\x03*\\\n\x15\x43ontrollerMessageType\x12\x0f\n\x0b\x43MT_GENERAL\x10\x00\x12\x0e\n\nCMT_ROTATE\x10\x01\x12\x11\n\rCMT_DISCOVERY\x10\x02\x12\x0f\n\x0b\x43MT_CONNECT\x10\x03*E\n\x10PAMRecordingType\x12\x0f\n\x0bPRT_SESSION\x10\x00\x12\x12\n\x0ePRT_TYPESCRIPT\x10\x01\x12\x0c\n\x08PRT_TIME\x10\x02\x42\x1f\n\x18\x63om.keepersecurity.protoB\x03PAMb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pam_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\003PAM'
  _globals['_WEBRTCCONNECTIONTYPE']._serialized_start=3236
  _globals['_WEBRTCCONNECTIONTYPE']._serialized_end=3378
  _globals['_PAMOPERATIONTYPE']._serialized_start=3380
  _globals['_PAMOPERATIONTYPE']._serialized_end=3444
  _globals['_PAMOPERATIONRESULTTYPE']._serialized_start=3446
  _globals['_PAMOPERATIONRESULTTYPE']._serialized_end=3558
  _globals['_CONTROLLERMESSAGETYPE']._serialized_start=3560
  _globals['_CONTROLLERMESSAGETYPE']._serialized_end=3652
  _globals['_PAMRECORDINGTYPE']._serialized_start=3654
  _globals['_PAMRECORDINGTYPE']._serialized_end=3723
  _globals['_PAMROTATIONSCHEDULE']._serialized_start=51
  _globals['_PAMROTATIONSCHEDULE']._serialized_end=182
  _globals['_PAMROTATIONSCHEDULESRESPONSE']._serialized_start=184
  _globals['_PAMROTATIONSCHEDULESRESPONSE']._serialized_end=259
  _globals['_PAMONLINECONTROLLER']._serialized_start=262
  _globals['_PAMONLINECONTROLLER']._serialized_end=410
  _globals['_PAMWEBRTCCONNECTION']._serialized_start=413
  _globals['_PAMWEBRTCCONNECTION']._serialized_end=580
  _globals['_PAMONLINECONTROLLERS']._serialized_start=582
  _globals['_PAMONLINECONTROLLERS']._serialized_end=671
  _globals['_PAMROTATEREQUEST']._serialized_start=673
  _globals['_PAMROTATEREQUEST']._serialized_end=730
  _globals['_PAMCONTROLLERSRESPONSE']._serialized_start=732
  _globals['_PAMCONTROLLERSRESPONSE']._serialized_end=797
  _globals['_PAMREMOVECONTROLLER']._serialized_start=799
  _globals['_PAMREMOVECONTROLLER']._serialized_end=860
  _globals['_PAMREMOVECONTROLLERRESPONSE']._serialized_start=862
  _globals['_PAMREMOVECONTROLLERRESPONSE']._serialized_end=938
  _globals['_PAMMODIFYREQUEST']._serialized_start=940
  _globals['_PAMMODIFYREQUEST']._serialized_end=1001
  _globals['_PAMDATAOPERATION']._serialized_start=1004
  _globals['_PAMDATAOPERATION']._serialized_end=1156
  _globals['_PAMCONFIGURATIONDATA']._serialized_start=1158
  _globals['_PAMCONFIGURATIONDATA']._serialized_end=1259
  _globals['_PAMELEMENTDATA']._serialized_start=1261
  _globals['_PAMELEMENTDATA']._serialized_end=1330
  _globals['_PAMELEMENTOPERATIONRESULT']._serialized_start=1332
  _globals['_PAMELEMENTOPERATIONRESULT']._serialized_end=1441
  _globals['_PAMMODIFYRESULT']._serialized_start=1443
  _globals['_PAMMODIFYRESULT']._serialized_end=1509
  _globals['_PAMELEMENT']._serialized_start=1511
  _globals['_PAMELEMENT']._serialized_end=1631
  _globals['_PAMGENERICUIDREQUEST']._serialized_start=1633
  _globals['_PAMGENERICUIDREQUEST']._serialized_end=1668
  _globals['_PAMGENERICUIDSREQUEST']._serialized_start=1670
  _globals['_PAMGENERICUIDSREQUEST']._serialized_end=1707
  _globals['_PAMCONFIGURATION']._serialized_start=1710
  _globals['_PAMCONFIGURATION']._serialized_end=1881
  _globals['_PAMCONFIGURATIONS']._serialized_start=1883
  _globals['_PAMCONFIGURATIONS']._serialized_end=1949
  _globals['_PAMCONTROLLER']._serialized_start=1952
  _globals['_PAMCONTROLLER']._serialized_end=2207
  _globals['_CONTROLLERRESPONSE']._serialized_start=2209
  _globals['_CONTROLLERRESPONSE']._serialized_end=2246
  _globals['_PAMCONFIGURATIONCONTROLLER']._serialized_start=2248
  _globals['_PAMCONFIGURATIONCONTROLLER']._serialized_end=2325
  _globals['_CONFIGURATIONADDREQUEST']._serialized_start=2328
  _globals['_CONFIGURATIONADDREQUEST']._serialized_end=2491
  _globals['_RELAYACCESSCREDS']._serialized_start=2493
  _globals['_RELAYACCESSCREDS']._serialized_end=2567
  _globals['_PAMRECORDING']._serialized_start=2570
  _globals['_PAMRECORDING']._serialized_end=2761
  _globals['_PAMRECORDINGSRESPONSE']._serialized_start=2763
  _globals['_PAMRECORDINGSRESPONSE']._serialized_end=2825
  _globals['_PAMDATA']._serialized_start=2827
  _globals['_PAMDATA']._serialized_end=2869
  _globals['_UIDLIST']._serialized_start=2871
  _globals['_UIDLIST']._serialized_end=2894
  _globals['_PAMRESOURCECONFIG']._serialized_start=2897
  _globals['_PAMRESOURCECONFIG']._serialized_end=3233
# @@protoc_insertion_point(module_scope)



from contextlib import contextmanager

from .attachment_reader import LastpassAttachmentReader
from ..importer import Attachment


class LastpassAttachment(Attachment):
    def __init__(self, id, parent, mimetype, storagekey, lastpass_size, filename):
        super().__init__()
        self.file_id = id
        self.parent = parent
        self.mime = mimetype
        self.storagekey = storagekey
        self.lastpass_size = lastpass_size
        self.name = filename
        self.tmpfile = None
        self.size = None  # Decrypted size to pass to importer
        self.key = None  # This lets the importer know to re-encrypt with a new Keeper key

    @contextmanager
    def open(self):
        with LastpassAttachmentReader.get_buffered_reader(self) as reader:
            yield reader

    @contextmanager
    def open_text(self):
        with LastpassAttachmentReader.get_text_reader(self, encoding='utf-8-sig') as reader:
            yield reader



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import logging
import os
import subprocess
import getpass
from contextlib import contextmanager
from typing import Dict, Any, Tuple, Optional

from fido2.webauthn import PublicKeyCredentialCreationOptions, PublicKeyCredentialRequestOptions

from .base import StorageHandler
from ..utils.constants import (
    ERROR_MESSAGES,
    WINDOWS_REGISTRY_PATH
)
from ..utils.error_handler import BiometricErrorHandler
from .base import BasePlatformHandler
import asyncio
from winrt.windows.security.credentials.ui import (
    UserConsentVerifier,
    UserConsentVerifierAvailability,
)


class WindowsStorageHandler(StorageHandler):
    """Windows Registry storage handler"""

    def __init__(self):
        self.key_path = WINDOWS_REGISTRY_PATH

    def _get_registry_key(self) -> Optional[Any]:
        """Get Windows registry key"""
        try:
            import winreg
            try:
                return winreg.OpenKey(winreg.HKEY_CURRENT_USER, self.key_path, 0, winreg.KEY_ALL_ACCESS)
            except FileNotFoundError:
                return winreg.CreateKey(winreg.HKEY_CURRENT_USER, self.key_path)
        except ImportError:
            return None

    @contextmanager
    def registry_key(self) -> Any:
        """Context manager for registry key handling"""
        import winreg
        key = None
        try:
            key = self._get_registry_key()
            if key is None:
                raise ImportError("winreg module not available")
            yield key
        finally:
            if key:
                winreg.CloseKey(key)

    def get_biometric_flag(self, username: str) -> bool:
        """Get biometric flag from Windows registry - True if credential ID exists"""
        return self.get_credential_id(username) is not None

    def delete_biometric_flag(self, username: str) -> bool:
        """Delete biometric flag from Windows registry - removes credential ID"""
        return self.delete_credential_id(username)

    def store_credential_id(self, username: str, credential_id: str) -> bool:
        """Store credential ID for user in Windows registry (also serves as biometric flag)"""
        try:
            import winreg
            with self.registry_key() as key:
                winreg.SetValueEx(key, username, 0, winreg.REG_SZ, credential_id)
                logging.debug("Stored credential ID for user: %s", username)
                return True
        except Exception as e:
            logging.warning("Failed to store credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("store credential ID", "Windows registry", e)
            return False

    def get_credential_id(self, username: str) -> Optional[str]:
        """Get stored credential ID for user from Windows registry"""
        try:
            import winreg
            with self.registry_key() as key:
                try:
                    value, reg_type = winreg.QueryValueEx(key, username)
                    if reg_type == winreg.REG_SZ and value:
                        logging.debug("Retrieved credential ID for user: %s", username)
                        return str(value)
                    else:
                        return None
                except FileNotFoundError:
                    logging.debug("No stored credential ID found for user: %s", username)
                    return None
        except Exception as e:
            logging.warning("Failed to retrieve credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("get credential ID", "Windows registry", e)
            return None

    def delete_credential_id(self, username: str) -> bool:
        """Delete stored credential ID for user from Windows registry"""
        try:
            import winreg
            with self.registry_key() as key:
                try:
                    winreg.DeleteValue(key, username)
                    logging.debug("Deleted stored credential ID for user: %s", username)
                    return True
                except FileNotFoundError:
                    logging.debug("Credential ID for user %s was already deleted", username)
                    return True
        except Exception as e:
            logging.warning("Failed to delete credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("delete credential ID", "Windows registry", e)
            return False


class WindowsHandler(BasePlatformHandler):
    """Windows-specific biometric handler"""

    def _create_storage_handler(self) -> StorageHandler:
        return WindowsStorageHandler()

    def _get_platform_name(self) -> str:
        return "Windows Hello"

    def _get_current_user_sid(self) -> Optional[str]:
        """Get current user SID using PowerShell and WMI"""
        try:
            cmd = ['powershell', '-Command', 
                   f"(Get-WmiObject -Class Win32_UserAccount -Filter \"Name='{getpass.getuser()}'\").SID"]
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
            return result.stdout.strip()
        except subprocess.CalledProcessError:
            try:
                cmd = ['whoami', '/user', '/fo', 'csv']
                result = subprocess.run(cmd, capture_output=True, text=True, check=True)
                lines = result.stdout.strip().split('\n')
                if len(lines) > 1:
                    sid_line = lines[1].split(',')
                    if len(sid_line) > 1:
                        return sid_line[1].strip('"')
            except subprocess.CalledProcessError:
                pass
            
            return None
        

    async def _check_biometrics(self) -> bool:
        """Check if biometrics (face/fingerprint) are enrolled using Windows Runtime API"""
        try:
            availability = await UserConsentVerifier.check_availability_async()

            if availability == UserConsentVerifierAvailability.AVAILABLE:
                return True
            else:
                return False
        except Exception as e:
            logging.debug("Failed to check biometrics availability: %s", str(e))
            return False

    # def _check_biometrics(self) -> bool:
    #     """Check if biometrics (face/fingerprint) are enrolled"""
    #     sid = self._get_current_user_sid()
    #     if not sid:
    #         return False
        
    #     reg_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\WinBio\AccountInfo\{}".format(sid)
    #     try:
    #         import winreg
    #         with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path) as key:
    #             value, regtype = winreg.QueryValueEx(key, "EnrolledFactors")
    #             # 2 = Face, 8 = Fingerprint, 10 = Face and Fingerprint
    #             return value in [2, 8, 10]
    #     except (FileNotFoundError, ImportError):
    #         return False

    # def _check_pin_enrollment(self) -> bool:
    #     """Check if PIN is enrolled"""
    #     sid = self._get_current_user_sid()
    #     if not sid:
    #         return False
        
    #     reg_path = r"SOFTWARE\Microsoft\Windows\CurrentVersion\Authentication\Credential Providers\{{D6886603-9D2F-4EB2-B667-1971041FA96B}}\{}".format(sid)
    #     try:
    #         import winreg
    #         with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, reg_path) as key:
    #             value, regtype = winreg.QueryValueEx(key, "LogonCredsAvailable")
    #             return value == 1
    #     except (FileNotFoundError, ImportError):
    #         return False

    def detect_capabilities(self) -> Tuple[bool, str]:
        """Detect Windows Hello capabilities"""
        if os.name != 'nt':
            return False, "Not running on Windows"

        try:
            # Run the async biometrics check
            has_biometrics = asyncio.run(self._check_biometrics())
            
            if has_biometrics:
                return True, "Windows Hello available: Biometrics"
            else:
                return False, ERROR_MESSAGES['windows_hello_not_setup']

        except Exception as e:
            logging.warning("Error detecting Windows Hello: %s", str(e))
            return False, f"Error detecting Windows Hello: {str(e)}"

    def create_webauthn_client(self, data_collector):
        """Create Windows WebAuthn client"""
        try:
            from fido2.client.windows import WindowsClient
            return WindowsClient(client_data_collector=data_collector)
        except ImportError:
            raise Exception('Windows Hello client not available. Install fido2[pcsc]')

    def handle_credential_creation(self, creation_options: Dict[str, Any]) -> Dict[str, Any]:
        """Handle Windows-specific credential creation"""
        return self._prepare_credential_creation_options(creation_options)

    def handle_authentication_options(self, pk_options: Dict[str, Any]) -> Dict[str, Any]:
        """Handle Windows-specific authentication options"""
        return self._prepare_authentication_options(pk_options)

    def perform_authentication(self, client, options: PublicKeyCredentialRequestOptions):
        """Perform Windows Hello authentication"""
        try:
            return client.get_assertion(options)
        except Exception as e:
            raise self._handle_authentication_error(e, self._get_platform_name())

    def perform_credential_creation(self, client, options: PublicKeyCredentialCreationOptions):
        """Perform Windows Hello credential creation"""
        try:
            return client.make_credential(options)
        except Exception as e:
            raise self._handle_credential_creation_error(e, self._get_platform_name()) 


import base64
import json
import logging
import os

from keeper_secrets_manager_core.utils import string_to_bytes, bytes_to_string

from ..folder import FolderMoveCommand
from ..record import RecordRemoveCommand
from ... import api, crypto, utils, vault, vault_extensions
from ...params import KeeperParams
from ...proto import pam_pb2, router_pb2


def pam_decrypt_configuration_data(pam_config_v6_record):
    pam_config_v6_record.get('data'),
    pam_config_v6_record.get('record_key_unencrypted')
    data_unencrypted_bytes = crypto.decrypt_aes_v2(
        utils.base64_url_decode(pam_config_v6_record.get('data')),
        pam_config_v6_record.get('record_key_unencrypted')
    )

    data_unencrypted_json_str = bytes_to_string(data_unencrypted_bytes)
    data_unencrypted_dict = json.loads(data_unencrypted_json_str)

    return data_unencrypted_dict


def pam_configuration_get_single_value_from_field(decrypted_record_dict, field_id):
    values = pam_configuration_get_all_values_from_field(decrypted_record_dict, field_id)
    if not values:
        return None

    if len(values) > 0:
        return values[0]

    return None


def pam_configuration_get_all_values_from_field(decrypted_record_dict, field_id):
    field = pam_configuration_get_field(decrypted_record_dict, field_id)
    if not field:
        return None

    return field.get('value')


def pam_configuration_get_field(unencrypted_record, field_identifier):

    if unencrypted_record is str:
        unencrypted_record = json.loads(unencrypted_record)

    fields = unencrypted_record.get('fields')

    if not fields:
        return None

    for field in fields:
        if field.get('id') == field_identifier:
            return field

        if field.get('type') == field_identifier:
            return field

        if field.get('label') == field_identifier:
            return field


def pam_configuration_create_record_v6(params, record, folder_uid):
    # type: (KeeperParams, vault.TypedRecord, str, str) -> None
    if not record.record_uid:
        record.record_uid = utils.generate_uid()

    if not record.record_key:
        record.record_key = utils.generate_aes_key()

    record_data = vault_extensions.extract_typed_record_data(record)
    json_data = api.get_record_data_json_bytes(record_data)

    car = pam_pb2.ConfigurationAddRequest()
    car.configurationUid = utils.base64_url_decode(record.record_uid)
    car.recordKey = crypto.encrypt_aes_v2(record.record_key, params.data_key)
    car.data = crypto.encrypt_aes_v2(json_data, record.record_key)

    api.communicate_rest(params, car, 'pam/add_configuration_record')


def pam_configuration_create(params, gateway_uid_bytes, config_json_str, child_config_json_strings=None, parent_uid_bytes=None):

    config_operation = pam_pb2.PAMDataOperation()
    config_operation.operationType = pam_pb2.PAMOperationType.ADD

    top_config_uid = os.urandom(16)

    if not parent_uid_bytes:
        # Root node
        config_operation.configuration.configurationUid = top_config_uid
        config_operation.configuration.controllerUid = gateway_uid_bytes
        config_operation.configuration.data = string_to_bytes(config_json_str)    # DATA size is between 16 and 516
    else:
        # Child node
        config_operation.element.elementUid = top_config_uid
        config_operation.element.parentUid = parent_uid_bytes
        config_operation.element.data = string_to_bytes(config_json_str)

    rq = pam_pb2.PAMModifyRequest()
    rq.operations.append(config_operation)

    child_generated_uids = []

    if child_config_json_strings:

        for child_config_json_string in child_config_json_strings:
            child_generated_uid = os.urandom(16)
            child_generated_uids.append(child_generated_uid)

            child_config_operation = pam_pb2.PAMDataOperation()
            child_config_operation.operationType = pam_pb2.PAMOperationType.ADD
            child_config_operation.element.elementUid = child_generated_uid
            child_config_operation.element.parentUid = top_config_uid
            child_config_operation.element.data = string_to_bytes(child_config_json_string)
            rq.operations.append(child_config_operation)


    rs = api.communicate_rest(params, rq, 'pam/modify_configuration', rs_type=pam_pb2.PAMModifyResult)

    return {
        'configUid': top_config_uid,
        'childConfigUids': child_generated_uids if child_generated_uids else None
    }


def config_update(params):
    pass


def pam_configuration_get_one(params, config_uid_urlsafe):
    # rq = PAMGenericUidRequest()
    # rq.uid = config_uid_bytes
    # rs = api.communicate_rest(params, rq, 'pam/get_configuration', rs_type=pam_pb2.PAMConfiguration)

    if config_uid_urlsafe not in params.record_cache:
        raise Exception(f'PAM Configuration record uid {config_uid_urlsafe} not found in cache. Make sure you are entering the correct uid.')
    config_record = params.record_cache[config_uid_urlsafe]

    data_unencrypted_bytes = crypto.decrypt_aes_v2(utils.base64_url_decode(config_record['data']), config_record['record_key_unencrypted'])
    data_unencrypted_json_str = bytes_to_string(data_unencrypted_bytes)
    data_unencrypted_dict = json.loads(data_unencrypted_json_str)
    config_record['data_decrypted'] = data_unencrypted_dict

    return config_record


def pam_configurations_get_all(params):

    # rs = api.communicate_rest(params, None, 'pam/get_configurations', rs_type=pam_pb2.PAMConfigurations)
    all_records = params.record_cache.values()
    all_v6_records = [rec for rec in list(all_records) if rec['version'] == 6]

    return all_v6_records


def pam_configuration_remove(params, configuration_uid):
    # TODO: Check if there are record rotations associated with this config and warn user about that before removing.
    RecordRemoveCommand().execute(params, record=configuration_uid, force=True)

    if configuration_uid in params.record_cache:
        del params.record_cache[configuration_uid]

    logging.info('PAM Configuration was removed successfully.')

    # raise Exception("Not implemented yet...")
    # # just do the regular record deletion
    # config_operation = PAMDataOperation()
    # config_operation.operationType = PAMOperationType.DELETE
    #
    # config_operation.configuration.configurationUid = configuration_uid
    #
    # rq = PAMModifyRequest()
    # rq.operations.append(config_operation)
    #
    # try:
    #     rs = api.communicate_rest(params, rq, 'pam/modify_configuration', rs_type=pam_pb2.PAMModifyResult)
    #     print(f'{bcolors.OKGREEN}PAM Configuration was removed successfully{bcolors.ENDC}')
    #
    # except KeeperApiError as ex:
    #     if ex.result_code == 'doesnt_exist':
    #         print(f'{bcolors.WARNING}This PAM Configuration does not exist{bcolors.ENDC}')
    #     else:
    #         print(f'{bcolors.WARNING}Error code: {ex.result_code}. {ex.message}{bcolors.ENDC}')


def record_rotation_get(params, record_uid_bytes):  # type: (KeeperParams, bytes) -> router_pb2.RouterRotationInfo

    rq = pam_pb2.PAMGenericUidRequest()
    rq.uid = record_uid_bytes

    rotation_info_rs = api.communicate_rest(params, rq, 'pam/get_rotation_info', rs_type=router_pb2.RouterRotationInfo)

    return rotation_info_rs


def configuration_controller_get(params, config_uid_bytes: bytes):
    """
    Get the Controller UID that has access to the configuration UID
    Retrieves a keeper.pam_controller record, from given configuration_uid provided in request.
    controller_uid is the UID of the user who has access to the configuration url_safe_str_to_bytes(config_uid)
    """
    rq = pam_pb2.PAMGenericUidRequest()
    rq.uid = config_uid_bytes

    config_info_rs = api.communicate_rest(params, rq, 'pam/get_configuration_controller',
                                          rs_type=pam_pb2.PAMController)

    return config_info_rs



from __future__ import annotations
from ...utils import value_to_boolean
import os
from typing import  TYPE_CHECKING

if TYPE_CHECKING:
    from ...params import KeeperParams
    from ...keeper_dag.connection import ConnectionBase


def get_connection(params: KeeperParams) -> ConnectionBase:
    if value_to_boolean(os.environ.get("USE_LOCAL_DAG", False)) is False:
        from ...keeper_dag.connection.commander import Connection as CommanderConnection
        return CommanderConnection(params=params)
    else:
        from ...keeper_dag.connection.local import Connection as LocalConnection
        return LocalConnection()


# -*- coding: utf-8 -*-
#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import logging
from typing import Dict

import requests
import msal     # pip install msal

from collections import namedtuple

from ... import vault
from ...commands.base import RecordMixin

AzureEndpoint = namedtuple('AzureEndpoint', 'activeDirectory microsoftGraphResourceId')

AZURE_GLOBAL_CLOUD = 'AzureCloud'
AZURE_CHINA_CLOUD = 'AzureChinaCloud'
AZURE_US_GOV_CLOUD = 'AzureUSGovernment'
AZURE_GERMAN_CLOUD = 'AzureGermanCloud'


AZURE_CLOUD_URLS = {
    AZURE_GLOBAL_CLOUD: AzureEndpoint('https://login.microsoftonline.com', 'https://graph.microsoft.com'),
    AZURE_CHINA_CLOUD: AzureEndpoint('https://login.chinacloudapi.cn', 'https://microsoftgraph.chinacloudapi.cn'),
    AZURE_US_GOV_CLOUD: AzureEndpoint('https://login.microsoftonline.us', 'https://graph.microsoft.us'),
    AZURE_GERMAN_CLOUD: AzureEndpoint('https://login.microsoftonline.de', 'https://graph.microsoft.de'),
}   # type: Dict[str, AzureEndpoint]


def rotate(record, newpassword):   # type: (vault.KeeperRecord, str) -> bool
    # The Azure user_id either as the object ID (GUID) or the user principal name (UPN) of the target user
    user_id = RecordMixin.get_record_field(record, 'login')

    tenant_id = RecordMixin.get_custom_field(record, "cmdr:azure_tenant_id")
    client_id = RecordMixin.get_custom_field(record, "cmdr:azure_client_id")
    secret = RecordMixin.get_custom_field(record, "cmdr:azure_secret")

    azure_cloud = RecordMixin.get_custom_field(record, "cmdr:azure_cloud")
    if isinstance(azure_cloud, str) and len(azure_cloud) > 0:
        azure_cloud = azure_cloud.lower()
        if 'china' in azure_cloud:
            azure_cloud = AZURE_CHINA_CLOUD
        elif 'usgov' in azure_cloud:
            azure_cloud = AZURE_US_GOV_CLOUD
        elif 'german' in azure_cloud:
            azure_cloud = AZURE_GERMAN_CLOUD
        elif 'global' in azure_cloud:
            azure_cloud = AZURE_GLOBAL_CLOUD
        elif 'azurecloud' == azure_cloud:
            azure_cloud = AZURE_GLOBAL_CLOUD
        else:
            logging.info('Cannot resolve Azure Cloud \"%s\". Default to Global Cloud.', azure_cloud)

    if azure_cloud not in AZURE_CLOUD_URLS:
        azure_cloud = AZURE_GLOBAL_CLOUD

    users_endpoint = f'{AZURE_CLOUD_URLS[azure_cloud].microsoftGraphResourceId}/v1.0/users'
    default_scope = f'{AZURE_CLOUD_URLS[azure_cloud].microsoftGraphResourceId}/.default'
    authority = f'{AZURE_CLOUD_URLS[azure_cloud].activeDirectory}/{tenant_id}'

    try:
        # Create a preferably long-lived app instance which maintains a token cache.
        app = msal.ConfidentialClientApplication(
            client_id,
            authority=authority,
            client_credential=secret,
            # token_cache=...  # Default cache is in memory only.
            # To learn how to use SerializableTokenCache from
            # https://msal-python.rtfd.io/en/latest/#msal.SerializableTokenCache
        )

        # Get access token
        result = app.acquire_token_silent(scopes=[default_scope], account=None)

        if not result:
            logging.debug("No suitable token exists in cache. Let's get a new one from AAD.")
            result = app.acquire_token_for_client(scopes=[default_scope])

        if "access_token" in result:
            access_token = result['access_token'] # JWT access token

            # 1. Getting all users from Azure Graph using the access token
            # all_users = requests.get(  # Use token to call downstream service
            #     users_endpoint,
            #     headers={'Authorization': 'Bearer ' + access_token}
            # ).json()
            #
            # 2. Getting only one user
            # usr = requests.get(
            #     '%s/%s' % (users_endpoint, user_id),
            #     headers={'Authorization': 'Bearer ' + access_token}
            # ).json()

            # 3. Updating user's password
            pwd_change_payload = {
                'passwordProfile': {
                    'password': newpassword,
                    'forceChangePasswordNextSignIn': False
                }
            }

            usr_pwd_update_resp = requests.patch(
                f'{users_endpoint}/{user_id}',
                headers={
                    'Authorization': 'Bearer ' + access_token,
                    'Content-Type': 'application/json'
                },
                data=json.dumps(pwd_change_payload)
            )

            resp_status_code = usr_pwd_update_resp.status_code

            if resp_status_code == 204:
                logging.info("Password successfully changed in Azure")
                return True

            elif resp_status_code == 403:
                resp_data = usr_pwd_update_resp.json()

                if resp_data['error']['code'] == 'Authorization_RequestDenied':
                    logging.error("Status code: %d, message: %s" % (resp_status_code, resp_data['error']['message']))
                    logging.error("Insufficient privileges to perform the password reset")
                    logging.error('\tIf you have access to Azure with administrator permission then you can enable\n'
                          '\tpermission by navigating to Azure Portal -> Azure AD -> Roles and administrators ->\n'
                          '\t"Helpdesk Administrator" -> Click on "Add Assignments" > select the application with client\n'
                          '\tid "%s" > click on Add button.' % client_id)
                else:
                    logging.error("Unknown status code: %d, message: %s" % (resp_status_code, resp_data['error']['message']))
            else:
                logging.error("Unhandled status code: %d, message: %s" % (resp_status_code, usr_pwd_update_resp.json()['error']['message']))
        else:
            logging.error('Azure error: %s, description: %s' % (result['error'], result['error_description']))
    except Exception as ex:
        logging.error(ex)

    return False



import APIRequest_pb2 as _APIRequest_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class AccountSummaryRequest(_message.Message):
    __slots__ = ["summaryVersion", "includeRecentActivity"]
    SUMMARYVERSION_FIELD_NUMBER: _ClassVar[int]
    INCLUDERECENTACTIVITY_FIELD_NUMBER: _ClassVar[int]
    summaryVersion: int
    includeRecentActivity: bool
    def __init__(self, summaryVersion: _Optional[int] = ..., includeRecentActivity: bool = ...) -> None: ...

class AccountSummaryElements(_message.Message):
    __slots__ = ["clientKey", "settings", "keysInfo", "syncLogs", "isEnterpriseAdmin", "license", "group", "Enforcements", "Images", "personalLicense", "fixSharedFolderRecords", "usernames", "devices", "isShareAdmin", "accountRecovery", "accountRecoveryPrompt", "minMasterPasswordLengthNoPrompt", "forbidKeyType2"]
    CLIENTKEY_FIELD_NUMBER: _ClassVar[int]
    SETTINGS_FIELD_NUMBER: _ClassVar[int]
    KEYSINFO_FIELD_NUMBER: _ClassVar[int]
    SYNCLOGS_FIELD_NUMBER: _ClassVar[int]
    ISENTERPRISEADMIN_FIELD_NUMBER: _ClassVar[int]
    LICENSE_FIELD_NUMBER: _ClassVar[int]
    GROUP_FIELD_NUMBER: _ClassVar[int]
    ENFORCEMENTS_FIELD_NUMBER: _ClassVar[int]
    IMAGES_FIELD_NUMBER: _ClassVar[int]
    PERSONALLICENSE_FIELD_NUMBER: _ClassVar[int]
    FIXSHAREDFOLDERRECORDS_FIELD_NUMBER: _ClassVar[int]
    USERNAMES_FIELD_NUMBER: _ClassVar[int]
    DEVICES_FIELD_NUMBER: _ClassVar[int]
    ISSHAREADMIN_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTRECOVERY_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTRECOVERYPROMPT_FIELD_NUMBER: _ClassVar[int]
    MINMASTERPASSWORDLENGTHNOPROMPT_FIELD_NUMBER: _ClassVar[int]
    FORBIDKEYTYPE2_FIELD_NUMBER: _ClassVar[int]
    clientKey: bytes
    settings: Settings
    keysInfo: KeysInfo
    syncLogs: _containers.RepeatedCompositeFieldContainer[SyncLog]
    isEnterpriseAdmin: bool
    license: License
    group: Group
    Enforcements: Enforcements
    Images: _containers.RepeatedCompositeFieldContainer[KeyValue]
    personalLicense: License
    fixSharedFolderRecords: bool
    usernames: _containers.RepeatedScalarFieldContainer[str]
    devices: _containers.RepeatedCompositeFieldContainer[DeviceInfo]
    isShareAdmin: bool
    accountRecovery: bool
    accountRecoveryPrompt: bool
    minMasterPasswordLengthNoPrompt: int
    forbidKeyType2: bool
    def __init__(self, clientKey: _Optional[bytes] = ..., settings: _Optional[_Union[Settings, _Mapping]] = ..., keysInfo: _Optional[_Union[KeysInfo, _Mapping]] = ..., syncLogs: _Optional[_Iterable[_Union[SyncLog, _Mapping]]] = ..., isEnterpriseAdmin: bool = ..., license: _Optional[_Union[License, _Mapping]] = ..., group: _Optional[_Union[Group, _Mapping]] = ..., Enforcements: _Optional[_Union[Enforcements, _Mapping]] = ..., Images: _Optional[_Iterable[_Union[KeyValue, _Mapping]]] = ..., personalLicense: _Optional[_Union[License, _Mapping]] = ..., fixSharedFolderRecords: bool = ..., usernames: _Optional[_Iterable[str]] = ..., devices: _Optional[_Iterable[_Union[DeviceInfo, _Mapping]]] = ..., isShareAdmin: bool = ..., accountRecovery: bool = ..., accountRecoveryPrompt: bool = ..., minMasterPasswordLengthNoPrompt: _Optional[int] = ..., forbidKeyType2: bool = ...) -> None: ...

class DeviceInfo(_message.Message):
    __slots__ = ["encryptedDeviceToken", "deviceName", "deviceStatus", "devicePublicKey", "encryptedDataKeyDoNotUse", "clientVersion", "username", "ipAddress", "approveRequestTime", "encryptedDataKeyPresent", "groupId", "devicePlatform", "clientFormFactor"]
    ENCRYPTEDDEVICETOKEN_FIELD_NUMBER: _ClassVar[int]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    DEVICESTATUS_FIELD_NUMBER: _ClassVar[int]
    DEVICEPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATAKEYDONOTUSE_FIELD_NUMBER: _ClassVar[int]
    CLIENTVERSION_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    APPROVEREQUESTTIME_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATAKEYPRESENT_FIELD_NUMBER: _ClassVar[int]
    GROUPID_FIELD_NUMBER: _ClassVar[int]
    DEVICEPLATFORM_FIELD_NUMBER: _ClassVar[int]
    CLIENTFORMFACTOR_FIELD_NUMBER: _ClassVar[int]
    encryptedDeviceToken: bytes
    deviceName: str
    deviceStatus: _APIRequest_pb2.DeviceStatus
    devicePublicKey: bytes
    encryptedDataKeyDoNotUse: bytes
    clientVersion: str
    username: str
    ipAddress: str
    approveRequestTime: int
    encryptedDataKeyPresent: bool
    groupId: int
    devicePlatform: str
    clientFormFactor: _APIRequest_pb2.ClientFormFactor
    def __init__(self, encryptedDeviceToken: _Optional[bytes] = ..., deviceName: _Optional[str] = ..., deviceStatus: _Optional[_Union[_APIRequest_pb2.DeviceStatus, str]] = ..., devicePublicKey: _Optional[bytes] = ..., encryptedDataKeyDoNotUse: _Optional[bytes] = ..., clientVersion: _Optional[str] = ..., username: _Optional[str] = ..., ipAddress: _Optional[str] = ..., approveRequestTime: _Optional[int] = ..., encryptedDataKeyPresent: bool = ..., groupId: _Optional[int] = ..., devicePlatform: _Optional[str] = ..., clientFormFactor: _Optional[_Union[_APIRequest_pb2.ClientFormFactor, str]] = ...) -> None: ...

class KeysInfo(_message.Message):
    __slots__ = ["encryptionParams", "encryptedDataKey", "dataKeyBackupDate", "userAuthUid", "encryptedPrivateKey", "encryptedEccPrivateKey", "eccPublicKey"]
    ENCRYPTIONPARAMS_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATAKEY_FIELD_NUMBER: _ClassVar[int]
    DATAKEYBACKUPDATE_FIELD_NUMBER: _ClassVar[int]
    USERAUTHUID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDECCPRIVATEKEY_FIELD_NUMBER: _ClassVar[int]
    ECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    encryptionParams: bytes
    encryptedDataKey: bytes
    dataKeyBackupDate: float
    userAuthUid: bytes
    encryptedPrivateKey: bytes
    encryptedEccPrivateKey: bytes
    eccPublicKey: bytes
    def __init__(self, encryptionParams: _Optional[bytes] = ..., encryptedDataKey: _Optional[bytes] = ..., dataKeyBackupDate: _Optional[float] = ..., userAuthUid: _Optional[bytes] = ..., encryptedPrivateKey: _Optional[bytes] = ..., encryptedEccPrivateKey: _Optional[bytes] = ..., eccPublicKey: _Optional[bytes] = ...) -> None: ...

class SyncLog(_message.Message):
    __slots__ = ["countryName", "secondsAgo", "deviceName", "countryCode", "deviceUID", "ipAddress"]
    COUNTRYNAME_FIELD_NUMBER: _ClassVar[int]
    SECONDSAGO_FIELD_NUMBER: _ClassVar[int]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    COUNTRYCODE_FIELD_NUMBER: _ClassVar[int]
    DEVICEUID_FIELD_NUMBER: _ClassVar[int]
    IPADDRESS_FIELD_NUMBER: _ClassVar[int]
    countryName: str
    secondsAgo: int
    deviceName: str
    countryCode: str
    deviceUID: bytes
    ipAddress: str
    def __init__(self, countryName: _Optional[str] = ..., secondsAgo: _Optional[int] = ..., deviceName: _Optional[str] = ..., countryCode: _Optional[str] = ..., deviceUID: _Optional[bytes] = ..., ipAddress: _Optional[str] = ...) -> None: ...

class License(_message.Message):
    __slots__ = ["subscriptionCode", "productTypeId", "productTypeName", "expirationDate", "secondsUntilExpiration", "maxDevices", "filePlanType", "bytesUsed", "bytesTotal", "secondsUntilStorageExpiration", "storageExpirationDate", "hasAutoRenewableAppstoreSubscription", "accountType", "uploadsRemaining", "enterpriseId", "chatEnabled", "auditAndReportingEnabled", "breachWatchFeatureDisable", "accountUid", "allowPersonalLicense", "licensedBy", "email", "breachWatchEnabled", "breachWatchScanned", "breachWatchExpiration", "breachWatchDateCreated", "error", "expiration", "storageExpiration", "uploadsCount", "units", "pendingEnterprise", "isPamEnabled"]
    SUBSCRIPTIONCODE_FIELD_NUMBER: _ClassVar[int]
    PRODUCTTYPEID_FIELD_NUMBER: _ClassVar[int]
    PRODUCTTYPENAME_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONDATE_FIELD_NUMBER: _ClassVar[int]
    SECONDSUNTILEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    MAXDEVICES_FIELD_NUMBER: _ClassVar[int]
    FILEPLANTYPE_FIELD_NUMBER: _ClassVar[int]
    BYTESUSED_FIELD_NUMBER: _ClassVar[int]
    BYTESTOTAL_FIELD_NUMBER: _ClassVar[int]
    SECONDSUNTILSTORAGEEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    STORAGEEXPIRATIONDATE_FIELD_NUMBER: _ClassVar[int]
    HASAUTORENEWABLEAPPSTORESUBSCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTTYPE_FIELD_NUMBER: _ClassVar[int]
    UPLOADSREMAINING_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEID_FIELD_NUMBER: _ClassVar[int]
    CHATENABLED_FIELD_NUMBER: _ClassVar[int]
    AUDITANDREPORTINGENABLED_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHFEATUREDISABLE_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTUID_FIELD_NUMBER: _ClassVar[int]
    ALLOWPERSONALLICENSE_FIELD_NUMBER: _ClassVar[int]
    LICENSEDBY_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHENABLED_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHSCANNED_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHDATECREATED_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    STORAGEEXPIRATION_FIELD_NUMBER: _ClassVar[int]
    UPLOADSCOUNT_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    PENDINGENTERPRISE_FIELD_NUMBER: _ClassVar[int]
    ISPAMENABLED_FIELD_NUMBER: _ClassVar[int]
    subscriptionCode: str
    productTypeId: int
    productTypeName: str
    expirationDate: str
    secondsUntilExpiration: int
    maxDevices: int
    filePlanType: int
    bytesUsed: int
    bytesTotal: int
    secondsUntilStorageExpiration: int
    storageExpirationDate: str
    hasAutoRenewableAppstoreSubscription: bool
    accountType: int
    uploadsRemaining: int
    enterpriseId: int
    chatEnabled: bool
    auditAndReportingEnabled: bool
    breachWatchFeatureDisable: bool
    accountUid: bytes
    allowPersonalLicense: bool
    licensedBy: str
    email: str
    breachWatchEnabled: bool
    breachWatchScanned: bool
    breachWatchExpiration: int
    breachWatchDateCreated: int
    error: Result
    expiration: int
    storageExpiration: int
    uploadsCount: int
    units: int
    pendingEnterprise: bool
    isPamEnabled: bool
    def __init__(self, subscriptionCode: _Optional[str] = ..., productTypeId: _Optional[int] = ..., productTypeName: _Optional[str] = ..., expirationDate: _Optional[str] = ..., secondsUntilExpiration: _Optional[int] = ..., maxDevices: _Optional[int] = ..., filePlanType: _Optional[int] = ..., bytesUsed: _Optional[int] = ..., bytesTotal: _Optional[int] = ..., secondsUntilStorageExpiration: _Optional[int] = ..., storageExpirationDate: _Optional[str] = ..., hasAutoRenewableAppstoreSubscription: bool = ..., accountType: _Optional[int] = ..., uploadsRemaining: _Optional[int] = ..., enterpriseId: _Optional[int] = ..., chatEnabled: bool = ..., auditAndReportingEnabled: bool = ..., breachWatchFeatureDisable: bool = ..., accountUid: _Optional[bytes] = ..., allowPersonalLicense: bool = ..., licensedBy: _Optional[str] = ..., email: _Optional[str] = ..., breachWatchEnabled: bool = ..., breachWatchScanned: bool = ..., breachWatchExpiration: _Optional[int] = ..., breachWatchDateCreated: _Optional[int] = ..., error: _Optional[_Union[Result, _Mapping]] = ..., expiration: _Optional[int] = ..., storageExpiration: _Optional[int] = ..., uploadsCount: _Optional[int] = ..., units: _Optional[int] = ..., pendingEnterprise: bool = ..., isPamEnabled: bool = ...) -> None: ...

class AddOn(_message.Message):
    __slots__ = ["licenseKeyId", "name", "expirationDate", "createdDate", "isTrial", "enabled", "scanned", "featureDisable"]
    LICENSEKEYID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONDATE_FIELD_NUMBER: _ClassVar[int]
    CREATEDDATE_FIELD_NUMBER: _ClassVar[int]
    ISTRIAL_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    SCANNED_FIELD_NUMBER: _ClassVar[int]
    FEATUREDISABLE_FIELD_NUMBER: _ClassVar[int]
    licenseKeyId: int
    name: str
    expirationDate: int
    createdDate: int
    isTrial: bool
    enabled: bool
    scanned: bool
    featureDisable: bool
    def __init__(self, licenseKeyId: _Optional[int] = ..., name: _Optional[str] = ..., expirationDate: _Optional[int] = ..., createdDate: _Optional[int] = ..., isTrial: bool = ..., enabled: bool = ..., scanned: bool = ..., featureDisable: bool = ...) -> None: ...

class Settings(_message.Message):
    __slots__ = ["audit", "mustPerformAccountShareBy", "shareAccountTo", "rules", "passwordRulesIntro", "autoBackupDays", "theme", "channel", "channelValue", "rsaConfigured", "emailVerified", "masterPasswordLastModified", "accountFolderKey", "securityKeys", "keyValues", "ssoUser", "onlineAccessOnly", "masterPasswordExpiry", "twoFactorRequired", "disallowExport", "restrictFiles", "restrictAllSharing", "restrictSharing", "restrictSharingIncomingAll", "restrictSharingIncomingEnterprise", "logoutTimer", "persistentLogin", "ipDisableAutoApprove", "shareDataKeyWithEccPublicKey", "shareDataKeyWithDevicePublicKey", "RecordTypesCounter", "RecordTypesEnterpriseCounter", "recordTypesEnabled", "canManageRecordTypes", "recordTypesPAMCounter", "logoutTimerMinutes", "securityKeysNoUserVerify", "channels", "personalUsernames"]
    AUDIT_FIELD_NUMBER: _ClassVar[int]
    MUSTPERFORMACCOUNTSHAREBY_FIELD_NUMBER: _ClassVar[int]
    SHAREACCOUNTTO_FIELD_NUMBER: _ClassVar[int]
    RULES_FIELD_NUMBER: _ClassVar[int]
    PASSWORDRULESINTRO_FIELD_NUMBER: _ClassVar[int]
    AUTOBACKUPDAYS_FIELD_NUMBER: _ClassVar[int]
    THEME_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    CHANNELVALUE_FIELD_NUMBER: _ClassVar[int]
    RSACONFIGURED_FIELD_NUMBER: _ClassVar[int]
    EMAILVERIFIED_FIELD_NUMBER: _ClassVar[int]
    MASTERPASSWORDLASTMODIFIED_FIELD_NUMBER: _ClassVar[int]
    ACCOUNTFOLDERKEY_FIELD_NUMBER: _ClassVar[int]
    SECURITYKEYS_FIELD_NUMBER: _ClassVar[int]
    KEYVALUES_FIELD_NUMBER: _ClassVar[int]
    SSOUSER_FIELD_NUMBER: _ClassVar[int]
    ONLINEACCESSONLY_FIELD_NUMBER: _ClassVar[int]
    MASTERPASSWORDEXPIRY_FIELD_NUMBER: _ClassVar[int]
    TWOFACTORREQUIRED_FIELD_NUMBER: _ClassVar[int]
    DISALLOWEXPORT_FIELD_NUMBER: _ClassVar[int]
    RESTRICTFILES_FIELD_NUMBER: _ClassVar[int]
    RESTRICTALLSHARING_FIELD_NUMBER: _ClassVar[int]
    RESTRICTSHARING_FIELD_NUMBER: _ClassVar[int]
    RESTRICTSHARINGINCOMINGALL_FIELD_NUMBER: _ClassVar[int]
    RESTRICTSHARINGINCOMINGENTERPRISE_FIELD_NUMBER: _ClassVar[int]
    LOGOUTTIMER_FIELD_NUMBER: _ClassVar[int]
    PERSISTENTLOGIN_FIELD_NUMBER: _ClassVar[int]
    IPDISABLEAUTOAPPROVE_FIELD_NUMBER: _ClassVar[int]
    SHAREDATAKEYWITHECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    SHAREDATAKEYWITHDEVICEPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    RECORDTYPESCOUNTER_FIELD_NUMBER: _ClassVar[int]
    RECORDTYPESENTERPRISECOUNTER_FIELD_NUMBER: _ClassVar[int]
    RECORDTYPESENABLED_FIELD_NUMBER: _ClassVar[int]
    CANMANAGERECORDTYPES_FIELD_NUMBER: _ClassVar[int]
    RECORDTYPESPAMCOUNTER_FIELD_NUMBER: _ClassVar[int]
    LOGOUTTIMERMINUTES_FIELD_NUMBER: _ClassVar[int]
    SECURITYKEYSNOUSERVERIFY_FIELD_NUMBER: _ClassVar[int]
    CHANNELS_FIELD_NUMBER: _ClassVar[int]
    PERSONALUSERNAMES_FIELD_NUMBER: _ClassVar[int]
    audit: bool
    mustPerformAccountShareBy: int
    shareAccountTo: _containers.RepeatedCompositeFieldContainer[MissingAccountShareKey]
    rules: _containers.RepeatedCompositeFieldContainer[PasswordRule]
    passwordRulesIntro: str
    autoBackupDays: int
    theme: str
    channel: str
    channelValue: str
    rsaConfigured: bool
    emailVerified: bool
    masterPasswordLastModified: float
    accountFolderKey: bytes
    securityKeys: _containers.RepeatedCompositeFieldContainer[SecurityKey]
    keyValues: _containers.RepeatedCompositeFieldContainer[KeyValue]
    ssoUser: bool
    onlineAccessOnly: bool
    masterPasswordExpiry: int
    twoFactorRequired: bool
    disallowExport: bool
    restrictFiles: bool
    restrictAllSharing: bool
    restrictSharing: bool
    restrictSharingIncomingAll: bool
    restrictSharingIncomingEnterprise: bool
    logoutTimer: int
    persistentLogin: bool
    ipDisableAutoApprove: bool
    shareDataKeyWithEccPublicKey: bool
    shareDataKeyWithDevicePublicKey: bool
    RecordTypesCounter: int
    RecordTypesEnterpriseCounter: int
    recordTypesEnabled: bool
    canManageRecordTypes: bool
    recordTypesPAMCounter: int
    logoutTimerMinutes: int
    securityKeysNoUserVerify: bool
    channels: _containers.RepeatedScalarFieldContainer[_APIRequest_pb2.TwoFactorChannelType]
    personalUsernames: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, audit: bool = ..., mustPerformAccountShareBy: _Optional[int] = ..., shareAccountTo: _Optional[_Iterable[_Union[MissingAccountShareKey, _Mapping]]] = ..., rules: _Optional[_Iterable[_Union[PasswordRule, _Mapping]]] = ..., passwordRulesIntro: _Optional[str] = ..., autoBackupDays: _Optional[int] = ..., theme: _Optional[str] = ..., channel: _Optional[str] = ..., channelValue: _Optional[str] = ..., rsaConfigured: bool = ..., emailVerified: bool = ..., masterPasswordLastModified: _Optional[float] = ..., accountFolderKey: _Optional[bytes] = ..., securityKeys: _Optional[_Iterable[_Union[SecurityKey, _Mapping]]] = ..., keyValues: _Optional[_Iterable[_Union[KeyValue, _Mapping]]] = ..., ssoUser: bool = ..., onlineAccessOnly: bool = ..., masterPasswordExpiry: _Optional[int] = ..., twoFactorRequired: bool = ..., disallowExport: bool = ..., restrictFiles: bool = ..., restrictAllSharing: bool = ..., restrictSharing: bool = ..., restrictSharingIncomingAll: bool = ..., restrictSharingIncomingEnterprise: bool = ..., logoutTimer: _Optional[int] = ..., persistentLogin: bool = ..., ipDisableAutoApprove: bool = ..., shareDataKeyWithEccPublicKey: bool = ..., shareDataKeyWithDevicePublicKey: bool = ..., RecordTypesCounter: _Optional[int] = ..., RecordTypesEnterpriseCounter: _Optional[int] = ..., recordTypesEnabled: bool = ..., canManageRecordTypes: bool = ..., recordTypesPAMCounter: _Optional[int] = ..., logoutTimerMinutes: _Optional[int] = ..., securityKeysNoUserVerify: bool = ..., channels: _Optional[_Iterable[_Union[_APIRequest_pb2.TwoFactorChannelType, str]]] = ..., personalUsernames: _Optional[_Iterable[str]] = ...) -> None: ...

class KeyValue(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: str
    def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class KeyValueBoolean(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: bool
    def __init__(self, key: _Optional[str] = ..., value: bool = ...) -> None: ...

class KeyValueLong(_message.Message):
    __slots__ = ["key", "value"]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    key: str
    value: int
    def __init__(self, key: _Optional[str] = ..., value: _Optional[int] = ...) -> None: ...

class Result(_message.Message):
    __slots__ = ["resultCode", "message", "result"]
    RESULTCODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    RESULT_FIELD_NUMBER: _ClassVar[int]
    resultCode: str
    message: str
    result: str
    def __init__(self, resultCode: _Optional[str] = ..., message: _Optional[str] = ..., result: _Optional[str] = ...) -> None: ...

class Enforcements(_message.Message):
    __slots__ = ["strings", "booleans", "longs", "jsons"]
    STRINGS_FIELD_NUMBER: _ClassVar[int]
    BOOLEANS_FIELD_NUMBER: _ClassVar[int]
    LONGS_FIELD_NUMBER: _ClassVar[int]
    JSONS_FIELD_NUMBER: _ClassVar[int]
    strings: _containers.RepeatedCompositeFieldContainer[KeyValue]
    booleans: _containers.RepeatedCompositeFieldContainer[KeyValueBoolean]
    longs: _containers.RepeatedCompositeFieldContainer[KeyValueLong]
    jsons: _containers.RepeatedCompositeFieldContainer[KeyValue]
    def __init__(self, strings: _Optional[_Iterable[_Union[KeyValue, _Mapping]]] = ..., booleans: _Optional[_Iterable[_Union[KeyValueBoolean, _Mapping]]] = ..., longs: _Optional[_Iterable[_Union[KeyValueLong, _Mapping]]] = ..., jsons: _Optional[_Iterable[_Union[KeyValue, _Mapping]]] = ...) -> None: ...

class MissingAccountShareKey(_message.Message):
    __slots__ = ["role_id", "publicKey"]
    ROLE_ID_FIELD_NUMBER: _ClassVar[int]
    PUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    role_id: int
    publicKey: bytes
    def __init__(self, role_id: _Optional[int] = ..., publicKey: _Optional[bytes] = ...) -> None: ...

class PasswordRule(_message.Message):
    __slots__ = ["ruleType", "pattern", "match", "minimum", "description", "value"]
    RULETYPE_FIELD_NUMBER: _ClassVar[int]
    PATTERN_FIELD_NUMBER: _ClassVar[int]
    MATCH_FIELD_NUMBER: _ClassVar[int]
    MINIMUM_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    ruleType: str
    pattern: str
    match: bool
    minimum: int
    description: str
    value: str
    def __init__(self, ruleType: _Optional[str] = ..., pattern: _Optional[str] = ..., match: bool = ..., minimum: _Optional[int] = ..., description: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class SecurityKey(_message.Message):
    __slots__ = ["deviceId", "deviceName", "dateAdded", "isValid", "deviceRegistration"]
    DEVICEID_FIELD_NUMBER: _ClassVar[int]
    DEVICENAME_FIELD_NUMBER: _ClassVar[int]
    DATEADDED_FIELD_NUMBER: _ClassVar[int]
    ISVALID_FIELD_NUMBER: _ClassVar[int]
    DEVICEREGISTRATION_FIELD_NUMBER: _ClassVar[int]
    deviceId: int
    deviceName: str
    dateAdded: int
    isValid: bool
    deviceRegistration: DeviceRegistration
    def __init__(self, deviceId: _Optional[int] = ..., deviceName: _Optional[str] = ..., dateAdded: _Optional[int] = ..., isValid: bool = ..., deviceRegistration: _Optional[_Union[DeviceRegistration, _Mapping]] = ...) -> None: ...

class DeviceRegistration(_message.Message):
    __slots__ = ["keyHandle", "publicKey", "attestationCert", "counter", "compromised"]
    KEYHANDLE_FIELD_NUMBER: _ClassVar[int]
    PUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    ATTESTATIONCERT_FIELD_NUMBER: _ClassVar[int]
    COUNTER_FIELD_NUMBER: _ClassVar[int]
    COMPROMISED_FIELD_NUMBER: _ClassVar[int]
    keyHandle: str
    publicKey: bytes
    attestationCert: str
    counter: int
    compromised: bool
    def __init__(self, keyHandle: _Optional[str] = ..., publicKey: _Optional[bytes] = ..., attestationCert: _Optional[str] = ..., counter: _Optional[int] = ..., compromised: bool = ...) -> None: ...

class Group(_message.Message):
    __slots__ = ["admin", "groupVerificationCode", "administrator"]
    ADMIN_FIELD_NUMBER: _ClassVar[int]
    GROUPVERIFICATIONCODE_FIELD_NUMBER: _ClassVar[int]
    ADMINISTRATOR_FIELD_NUMBER: _ClassVar[int]
    admin: bool
    groupVerificationCode: str
    administrator: Administrator
    def __init__(self, admin: bool = ..., groupVerificationCode: _Optional[str] = ..., administrator: _Optional[_Union[Administrator, _Mapping]] = ...) -> None: ...

class Administrator(_message.Message):
    __slots__ = ["firstName", "lastName", "email", "currentNumberOfUsers", "numberOfUsers", "subscriptionCode", "expirationDate", "purchaseDate"]
    FIRSTNAME_FIELD_NUMBER: _ClassVar[int]
    LASTNAME_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    CURRENTNUMBEROFUSERS_FIELD_NUMBER: _ClassVar[int]
    NUMBEROFUSERS_FIELD_NUMBER: _ClassVar[int]
    SUBSCRIPTIONCODE_FIELD_NUMBER: _ClassVar[int]
    EXPIRATIONDATE_FIELD_NUMBER: _ClassVar[int]
    PURCHASEDATE_FIELD_NUMBER: _ClassVar[int]
    firstName: str
    lastName: str
    email: str
    currentNumberOfUsers: int
    numberOfUsers: int
    subscriptionCode: str
    expirationDate: str
    purchaseDate: str
    def __init__(self, firstName: _Optional[str] = ..., lastName: _Optional[str] = ..., email: _Optional[str] = ..., currentNumberOfUsers: _Optional[int] = ..., numberOfUsers: _Optional[int] = ..., subscriptionCode: _Optional[str] = ..., expirationDate: _Optional[str] = ..., purchaseDate: _Optional[str] = ...) -> None: ...



from __future__ import annotations
import json
from ..pam.router_helper import router_send_action_to_gateway
from ..pam.pam_dto import GatewayAction
from ..pam.router_helper import get_response_payload
from ...proto import pam_pb2
from ...display import bcolors
from ... import vault
from ...discovery_common.record_link import RecordLink
from ... import utils
import logging
import requests
import hmac
import hashlib
import os
from pydantic import BaseModel
from typing import Optional, List, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..discover import GatewayContext
    from ...params import KeeperParams
    from ...vault import TypedRecord
    from ...keeper_dag.vertex import DAGVertex


CATALOG_REPO = "Keeper-Security/discovery-and-rotation-saas-dev"


class GatewayActionSaasListCommandInputs:

    def __init__(self,
                 configuration_uid: str,
                 languages: Optional[List[str]] = None):

        if languages is None:
            languages = ["en_US"]

        self.configurationUid = configuration_uid
        self.languages = languages

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)


class GatewayActionSaasListCommand(GatewayAction):

    def __init__(self, inputs: GatewayActionSaasListCommandInputs, conversation_id=None):
        super().__init__('saas-list', inputs=inputs, conversation_id=conversation_id, is_scheduled=True)

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)


# These are from KDNRM saas_type.py
class SaasConfigEnum(BaseModel):
    value: str
    desc: Optional[str] = None
    code: Optional[str] = None


class SaasConfigItem(BaseModel):
    id: str
    label: str
    desc: str
    is_secret: bool = False
    desc_code: Optional[str] = None
    type: Optional[str] = "text"
    code: Optional[str] = None
    default_value: Optional[Any] = None
    enum_values: List[SaasConfigEnum] = []
    required: bool = False


class SaasPluginUsage(BaseModel):
    record_id: str
    plugin_name: str
    user_uids: List[str] = []


class SaasCatalog(BaseModel):
    name: str
    type: str = "catalog"
    author: Optional[str] = None
    email: Optional[str] = None
    summary: Optional[str] = None
    file: Optional[str] = None
    file_sig: Optional[str] = None
    allows_remote_management: Optional[bool] = False
    readme: Optional[str] = None
    fields: List[SaasConfigItem] = []
    installed: bool = False
    used_by: List[SaasPluginUsage] = []

    @property
    def file_name(self):
        return self.file.split(os.sep)[-1] if self.file else None


def get_gateway_saas_schema(params: KeeperParams, gateway_context: GatewayContext) -> Optional[List[dict]]:

    """
    Get a plugins list from the Gateway.

    Using the plugins from the Gateway handles problem with versions.
    We can work off the builtin, and custom, plugins available to the version of the Gateway.
    """

    if gateway_context is None:
        print(f"{bcolors.FAIL}The user record does not have the set gateway{bcolors.ENDC}")
        return None

    # Get schema information from the Gateway
    action_inputs = GatewayActionSaasListCommandInputs(
        configuration_uid=gateway_context.configuration_uid,
    )

    conversation_id = GatewayAction.generate_conversation_id()
    router_response = router_send_action_to_gateway(
        params=params,
        gateway_action=GatewayActionSaasListCommand(
            inputs=action_inputs,
            conversation_id=conversation_id),
        message_type=pam_pb2.CMT_GENERAL,
        is_streaming=False,
        destination_gateway_uid_str=gateway_context.gateway_uid
    )

    if router_response is None:
        print(f"{bcolors.FAIL}Did not get router response.{bcolors.ENDC}")
        return None

    response = router_response.get("response")
    logging.debug(f"Router Response: {response}")
    payload = get_response_payload(router_response)
    data = payload.get("data")
    if data is None:
        raise Exception("The router returned a failure.")
    elif data.get("success") is False:
        error = data.get("error")
        logging.debug(f"gateway returned: {error}")
        print(f"{bcolors.FAIL}Could not get a list of SaaS plugins available on the gateway.{bcolors.ENDC}")
        return None

    return data.get("data", [])


def find_user_saas_configurations(params: KeeperParams, gateway_context: GatewayContext) -> dict:

    """
    Find all the SaaS configuration being uses by a gateway.



    """

    record_linking = RecordLink(record=gateway_context.configuration,  params=params, logger=logging)

    def _walk_graph(v: DAGVertex, m: dict, pv: Optional[DAGVertex] = None):

        # Skip any disabled vertices
        if not v.active:
            return

        # Get the record for this vertex; skip is not a pamUser.
        record = vault.TypedRecord.load(params, v.uid)
        if record is not None and record.record_type == "pamUser" and pv is not None:
            acl = record_linking.get_acl(v.uid, pv.uid)
            if acl is not None and acl.rotation_settings is not None:
                config_record_uid_list = acl.rotation_settings.saas_record_uid_list
                if config_record_uid_list is not None:
                    for config_record_uid in config_record_uid_list:

                        if config_record_uid not in m:
                            config_record = vault.TypedRecord.load(params,
                                                                   config_record_uid)  # type: Optional[TypedRecord]
                            if config_record is None:
                                continue

                            plugin_name = next((f.value for f in config_record.custom if f.label == "SaaS Type"),
                                               None)
                            if plugin_name is None:
                                continue

                            m[config_record_uid] = SaasPluginUsage(
                                record_id=config_record_uid,
                                plugin_name=plugin_name[0]
                            )  # type: SaasPluginUsage

                        m[config_record_uid].user_uids.append(v.uid)

        for next_v in v.has_vertices():
            _walk_graph(
                v=next_v,
                pv=v,
                m=m
            )

    usage_map = {}
    _walk_graph(
        v=record_linking.dag.get_root,
        m=usage_map)

    return usage_map


def get_plugins_map(params: KeeperParams, gateway_context: GatewayContext) -> Optional[dict[str, SaasCatalog]]:

    """
    Get a map of all the available plugins.

    This will first get the latest catalog from the GitHub repo.
    The catalog will contain the plugin available from the repo and built in.

    Then the Gateway is checked for custom plugin; plugins outside our control.

    The result is a dictionary, with the plugin name as the key.

    """

    plugin_map = {}

    # #### GATEWAY PLUGINS

    # Get a list of installed plugins (custom and builtin) from the Gateway
    gateway_plugins = get_gateway_saas_schema(params, gateway_context)
    if gateway_plugins is None:
        return None

    # Add the Gateway plugins to map; all these plugins are installed.
    for plugin_dict in gateway_plugins:
        plugin = SaasCatalog.model_validate(plugin_dict)  # type: SaasCatalog
        plugin.installed = True
        plugin_map[plugin.name] = plugin

    # #### CATALOG PLUGINS

    # Get the latest release of the catalog.json
    api_url = f"https://api.github.com/repos/{CATALOG_REPO}/releases/latest"
    res = utils.ssl_aware_get(api_url)
    if res.ok is False:
        print("")
        print(f"{bcolors.FAIL}Could not get plugin catalog from GitHub.{bcolors.ENDC}")
        return None
    release_data = res.json()

    # Find the latest release URL
    assets = release_data.get("assets", [])
    asset = assets[0]
    download_url = asset["browser_download_url"]
    logging.debug(f"download {asset['name']} from {download_url}")

    # Download the latest the catalog.yml
    res = utils.ssl_aware_get(download_url)
    if res.ok is False:
        print("")
        print(f"{bcolors.FAIL}Could not download the plugin catalog from GitHub.{bcolors.ENDC}")
        return None

    # Get a mapping of all the plugins being used by the plugin name.
    # The group usage by plugin name; we can have multiple configuration for the same plugin
    # This return dictionary of config record UID to SaasPluginUsage
    plugin_usage = find_user_saas_configurations(params, gateway_context)
    plugin_usage_map = {}
    for config_record_uid in plugin_usage:  # type: str
        plugin_name = plugin_usage[config_record_uid].plugin_name
        if plugin_name not in plugin_usage_map:
            plugin_usage_map[plugin_name] = []
        plugin_usage_map[plugin_name].append(plugin_usage[config_record_uid])

    for plugin_dict in json.loads(res.content):  # type: dict
        if plugin_dict.get("type") == "builtin":
            continue

        plugin = SaasCatalog.model_validate(plugin_dict)  # type: SaasCatalog
        if plugin.name in plugin_map:
            logging.debug(f"found duplicate plugin {plugin.name}; using plugin from gateway.")
            continue
        plugin_map[plugin.name] = plugin

    return plugin_map


def make_script_signature(plugin_code_bytes: bytes) -> str:

    # To use HMAC, we need to have a key; the key is not a secret, we just want to make a unique digest.
    this_is_not_a_secret = b"NOT_IMPORTANT"
    return hmac.new(this_is_not_a_secret, plugin_code_bytes, hashlib.sha256).hexdigest()


def get_field_input(field, current_value: Optional[str] = None):

    logging.debug(field.model_dump_json())

    print(f"{bcolors.BOLD}{field.label}{bcolors.ENDC}")
    print(f"Description: {field.desc}")
    if field.required is True:
        print(f"{bcolors.WARNING}Field is required.{bcolors.ENDC}")
    if field.type == "multiline":
        print(f"Enter a file path to load value from file.")

    while True:
        prompt = "Enter value"
        extra_text = []
        valid_values = []
        if len(field.enum_values) > 0:
            valid_values = [str(x.value) for x in field.enum_values]
            extra_text.append(f"Allowed values: {bcolors.BOLD}" +
                              f"{bcolors.ENDC}, {bcolors.BOLD}".join(valid_values) + bcolors.ENDC)
        if current_value is not None:
            extra_text.append(f"Enter for current value '{bcolors.OKGREEN}{current_value}{bcolors.ENDC}'")
        if field.default_value is not None:
            extra_text.append(f"Enter for default value '{bcolors.OKBLUE}{field.default_value}{bcolors.ENDC}'")
        if len(extra_text) > 0:
            prompt += f" (" + "; ".join(extra_text) + ")"
        prompt += " > "
        value = input(prompt)
        if value == "":
            if current_value is not None:
                value = current_value
            elif field.default_value is not None:
                value = field.default_value
        elif os.path.exists(value):
            with open(value, "r") as fh:
                value = fh.read()
                fh.close()
        if len(valid_values) > 0 and value not in valid_values:
            print(f"{bcolors.FAIL}{value} is not a valid value.{bcolors.ENDC}")
            continue
        if value is not None:
            break
        if field.required is False:
            break

        print(f"{bcolors.FAIL}This field is required.{bcolors.ENDC}")

    return [value]


def get_record_field_value(record: TypedRecord, label: str) -> Optional[str]:

    field = next((f for f in record.custom if f.label == label), None)
    if field is None or field.value is None or len(field.value) == 0 or field.value[0] is None or field.value[0] == "":
        return None
    return field.value[0]


def set_record_field_value(record: TypedRecord, label: str, value: str, field_type: Optional[str] = "text"):

    if value is not None and isinstance(value, list):
        value = value[0]

    field = next((f for f in record.custom if f.label == label), None)
    if field is None or field.value is None or len(field.value) == 0 or field.value[0] is None or field.value[0] == "":
        if value is not None:
            record.custom.append(
                vault.TypedField.new_field(
                    field_label=label,
                    field_type=field_type,
                    field_value=[value]
                )
            )
    elif value is not None:
        field.value = [value]
    else:
        field.value = []
